/*
** Module   : SaxWriterImpl
** Abstract : SAX writer specific features implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130208 Created initial version. Implementation is limited to the stubs to get the
**                  converted java code compiled.
** 002 CA  20130221 This resource doesn't have NAME attribute support.
** 003 EVL 20130320 Spelling fixes for comments.
** 004 EVK 20131001 Implement methods and class logic.
*/

package com.goldencode.p2j.xml;

import com.goldencode.p2j.util.*;
import javanet.staxutils.IndentingXMLStreamWriter;
import org.codehaus.stax2.XMLOutputFactory2;
import org.w3c.dom.Node;
import org.xml.sax.Attributes;

import javax.xml.stream.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Logger;

/**
 * Implementation SAX-Writer specific features of the SAX objects.
 */
public class SaxWriterImpl
extends SaxEntityImpl
implements SaxWriter
{
   /** default version of XML */
   private static final String DEFAULT_XML_VERSION = "1.0";

   /** Anonymous log instance. */
   private static final Logger LOG = LogHelper.getLogger(SaxWriterImpl.class.getName());

   /** Current status of SAX-Writer */
   private WriteStatus writeStatus = WriteStatus.SAX_WRITE_IDLE;

   /** Current output source of SAX-Writer */
   private OutputStream outputWrapper = null;

   /** encoding of SAX-Writer */
   private String encoding = "";

   /** version of xml */
   private String version = DEFAULT_XML_VERSION;

   /** formatted attribute */
   private boolean formatted = false;

   /** strict attribute */
   private boolean strict = true;

   /** fragment attribute */
   private boolean fragment = false;

   /** standalone attribute */
   private logical standalone = new logical();

   /** XML stream cursor writer*/
   private XMLStreamWriter writer = null;

   /**
    * Default constructor.
    */
   public SaxWriterImpl()
   {
      // no-op
   }

   /**
    * Gets the value of the encoding attribute for the given SAX-Writer object.
    *
    * @return The character value of the encoding attribute.
    */
   @Override
   public character getEncoding()
   {
      return new character(encoding);
   }

   /**
    * Sets the new value of the encoding attribute for the given SAX-Writer object.
    *
    * @param    encoding
    *           The encoding value to be set for the given writer.
    */
   @Override
   public void setEncoding(String encoding)
   {
      setEncoding(new character(encoding));
   }

   /**
    * Sets the new value of the encoding attribute for the given SAX-Writer object.
    *
    * @param    encoding
    *           The encoding value to be set for the given writer.
    */
   @Override
   public void setEncoding(character encoding)
   {
      if (encoding.isUnknown())
      {
         invalidArgumentAssignError("ENCODING");
         return;
      }

      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusAttributeError("ENCODING", writeStatus);
         return;
      }

      this.encoding = encoding.toStringMessage();
   }

   /**
    * Gets the value of the formatted attribute for the given SAX-Writer object. Formatted means
    * extra space or carrage returns should be added to preserve the hierarchical view of the
    * document.
    * 
    * @return   The <code>true</code> in case of formatted <code>false</code> otherwise.
    */
   @Override
   public logical getFormatted()
   {
      return new logical(formatted);
   }
   
   /**
    * Sets the value of the formatted attribute for the given SAX-Writer object. Formatted means
    * extra space or carriage returns should be added to preserve the hierarchical view of the
    * document.
    * 
    * @param    isFormatted
    *           The <code>true</code> in case of formatted <code>false</code> otherwise.
    */
   @Override
   public void setFormatted(logical isFormatted)
   {
      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusAttributeError("FORMATTED", writeStatus);
         return;
      }

      formatted = isFormatted.booleanValue();
   }

   /**
    * Sets the value of the formatted attribute for the given SAX-Writer object. Formatted means
    * extra space or carrage returns should be added to preserve the hierarchical view of the
    * document.
    *
    * @param    isFormatted
    *           The <code>true</code> in case of formatted <code>false</code> otherwise.
    */
   @Override
   public void setFormatted(boolean isFormatted)
   {
      setFormatted(new logical(isFormatted));
   }

   /**
    * Gets the value of the strict attribute for the given SAX-Writer object. Strict means
    * SAX-Writer should make well formed XML document. The default is true.
    * 
    * @return   The <code>true</code> in case of strict <code>false</code> otherwise.
    */
   @Override
   public logical getStrict()
   {
      return new logical(strict);
   }
   
   /**
    * Sets the value of the strict attribute for the given SAX-Writer object. Strict means
    * SAX-Writer should make well formed XML document. The default is true.
    * 
    * @param    isStrict
    *           The <code>true</code> in case of strict <code>false</code> otherwise.
    */
   @Override
   public void setStrict(logical isStrict)
   {
      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusAttributeError("STRICT", writeStatus);
         return;
      }

      strict = isStrict.booleanValue();
   }

   /**
    * Sets the value of the strict attribute for the given SAX-Writer object. Strict means
    * SAX-Writer should make well formed XML document. The default is true.
    * 
    * @param    isStrict
    *           The <code>true</code> in case of strict <code>false</code> otherwise.
    */
   @Override
   public void setStrict(boolean isStrict)
   {
      setStrict(new logical(isStrict));
   }

   /**
    * Gets the value of the fragment attribute for the given SAX-Writer object. Fragment means
    * SAX-Writer output is part of the document or full.
    * 
    * @return   The <code>true</code> in case of fragment <code>false</code> for full document.
    */
   @Override
   public logical getFragment()
   {
      return new logical(fragment);
   }
   
   /**
    * Sets the value of the fragment attribute for the given SAX-Writer object. Fragment means
    * SAX-Writer output is part of the document or full.
    * 
    * @param    isFragment
    *           The <code>true</code> in case of fragment <code>false</code> for full document.
    */
   @Override
   public void setFragment(logical isFragment)
   {
      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusAttributeError("FRAGMENT", writeStatus);
         return;
      }

      fragment = isFragment.booleanValue();
   }
   
   /**
    * Sets the value of the fragment attribute for the given SAX-Writer object. Fragment means
    * SAX-Writer output is part of the document or full.
    * 
    * @param    isFragment
    *           The <code>true</code> in case of fragment <code>false</code> for full document.
    */
   @Override
   public void setFragment(boolean isFragment)
   {
      setFragment(new logical(isFragment));
   }
   
   /**
    * Gets the value of the standalone attribute for the given SAX-Writer object.
    * 
    * @return   The character value of the standalone attribute.
    */
   @Override
   public logical getStandalone()
   {
      return new logical(standalone);
   }
   
   /**
    * Sets the value of the standalone attribute for the given SAX-Writer object.
    * 
    * @param    isStandalone
    *           The <code>true</code> in case of standalone <code>false</code> otherwise.
    */
   @Override
   public void setStandalone(logical isStandalone)
   {
      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusAttributeError("STANDALONE", writeStatus);
         return;
      }

      standalone = new logical(isStandalone);
   }
   
   /**
    * Sets the value of the standalone attribute for the given SAX-Writer object.
    * 
    * @param    isStandalone
    *           The <code>true</code> in case of standalone <code>false</code> otherwise.
    */
   @Override
   public void setStandalone(boolean isStandalone)
   {
      setStandalone(new logical(isStandalone));
   }
   
   /**
    * Gets the value of the version attribute for the given SAX-Writer object.
    * 
    * @return   The character value of the version attribute.
    */
   @Override
   public character getVersion()
   {
      return new character(version);
   }
   
   /**
    * Sets the new value of the version attribute for the given SAX-Writer object.
    * 
    * @param    version
    *           The version value to be set for the given writer.
    */
   @Override
   public void setVersion(String version)
   {
      setVersion(new character(version));
   }
   
   /**
    * Sets the new value of the version attribute for the given SAX-Writer object.
    * 
    * @param    version
    *           The version value to be set for the given writer.
    */
   @Override
   public void setVersion(character version)
   {
      if (version.isUnknown())
      {
         invalidArgumentAssignError("VERSION");
         return;
      }

      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusAttributeError("VERSION", writeStatus);
         return;
      }

      this.version = version.toStringMessage();
   }
   
   /**
    * Gets the value of the write-status attribute for the given SAX-Writer object. This is one
    * of the SAX_WRITE_* constants defined by the {@link SaxWriter} interface.
    * <p>
    * Documentation states that the type of this attribute is character, but this is incorrect. 
    * The attribute's real type is integer, same as PARSE-STATUS for the SAX-Reader object.
    * 
    * @return   The integer value of the write-status attribute.
    */
   @Override
   public integer getWriteStatus()
   {
      return new integer(writeStatus.getStatus());
   }
   
   /**
    * Adds namespace declaration to a tag in a XML document associated with SAX-Writer object.
    * 
    * @param    uri
    *           The URI namespace to add.
    * @param    prefix
    *           The prefix of the namespace to add.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical declareNamespace(Text uri, Text prefix)
   {
      if (!WriteStatus.writableBodyStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("DECLARE-NAMESPACE", writeStatus);
         return new logical(false);
      }

      String prefixStr = prefix.isUnknown() ? null : prefix.toStringMessage();
      try
      {
         writer.writeNamespace(prefixStr, uri.toStringMessage());
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("DECLARE-NAMESPACE");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Adds namespace declaration to a tag in a XML document associated with SAX-Writer object.
    *
    * @param    uri
    *           The URI namespace to add.
    * @param    prefix
    *           The prefix of the namespace to add.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical declareNamespace(String uri, Text prefix)
   {
      return declareNamespace(new character(uri), prefix);
   }

   /**
    * Adds namespace declaration to a tag in a XML document associated with SAX-Writer object.
    * 
    * @param    uri
    *           The URI namespace to add.
    * @param    prefix
    *           The prefix of the namespace to add.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical declareNamespace(Text uri, String prefix)
   {
      return declareNamespace(uri, new character(prefix));
   }

   /**
    * Adds namespace declaration to a tag in a XML document associated with SAX-Writer object.
    * 
    * @param    uri
    *           The URI namespace to add.
    * @param    prefix
    *           The prefix of the namespace to add.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical declareNamespace(String uri, String prefix)
   {
      return declareNamespace(new character(uri), new character(prefix));
   }

   /**
    * Adds namespace declaration to a tag in a XML document associated with SAX-Writer object.
    * 
    * @param    uri
    *           The URI namespace to add.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical declareNamespace(String uri)
   {
      return declareNamespace(new character(uri), new character());
   }
   
   /**
    * Adds namespace declaration to a tag in a XML document associated with SAX-Writer object.
    * 
    * @param    uri
    *           The URI namespace to add.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical declareNamespace(Text uri)
   {
      return declareNamespace(uri, new character());
   }
   
   /**
    * Closes the XML document associated with SAX-Writer object.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical endDocument()
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("END-DOCUMENT", writeStatus);
         return new logical(false);
      }

      try
      {
         writer.writeEndDocument();
         writeStatus = WriteStatus.SAX_WRITE_COMPLETE;
         outputWrapper.close();
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("END-DOCUMENT");
         return new logical(false);
      }
      catch (IOException ignored)
      {
         writerOutputError("END-DOCUMENT");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Ends an XML node of the given name in a XML document associated with SAX-Writer object.
    * 
    * @param    name
    *           The name of the node to end.
    * @param    uri
    *           The URI namespace of the node to end.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical endElement(Text name, Text uri)
   {
      if (!BaseDataType.elementsOfType(character.class, uri, name))
      {
         invalidArgumentError("END-ELEMENT");
         return new logical(false);
      }

      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("END-ELEMENT", writeStatus);
         return new logical(false);
      }

      try
      {
         writer.writeEndElement();
         writeStatus = WriteStatus.SAX_WRITE_ELEMENT;
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("END-ELEMENT");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Ends an XML node of the given name in a XML document associated with SAX-Writer object.
    * 
    * @param    name
    *           The name of the node to end.
    * @param    uri
    *           The URI namespace of the node to end.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical endElement(String name, Text uri)
   {
      return endElement(new character(name), uri);
   }

   /**
    * Ends an XML node of the given name in a XML document associated with SAX-Writer object.
    * 
    * @param    name
    *           The name of the node to end.
    * @param    uri
    *           The URI namespace of the node to end.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical endElement(Text name, String uri)
   {
      return endElement(name, new character(uri));
   }

   /**
    * Ends an XML node of the given name in a XML document associated with SAX-Writer object.
    * 
    * @param    name
    *           The name of the node to end.
    * @param    uri
    *           The URI namespace of the node to end.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical endElement(String name, String uri)
   {
      return endElement(new character(name), new character(uri));
   }

   /**
    * Ends an XML node of the given name in a XML document associated with SAX-Writer object.
    * 
    * @param    name
    *           The name of the node to end.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical endElement(String name)
   {
      return endElement(new character(name), new character());
   }
   
   /**
    * Ends an XML node of the given name in a XML document associated with SAX-Writer object.
    * 
    * @param    name
    *           The name of the node to end.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical endElement(Text name)
   {
      return endElement(name, new character());
   }
   
   /**
    * Specifies the output destination of the XML document to be created by a SAX-Writer object.
    * 
    * @param    mode
    *           The type of the destination to write to. Can be 'FILE', 'STREAM', 'MEMPTR', 
    *           'STREAM-HANDLE' or 'LONGCHAR'.
    * @param    destination
    *           The appropriate object to write.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setOutputDestination(String mode, Object destination)
   {
      return setOutputDestination(new character(mode), destination);
   }

   /**
    * Specifies the output destination of the XML document to be created by a SAX-Writer object.
    * 
    * @param    mode
    *           The type of the destination to write to. Can be 'FILE', 'STREAM', 'MEMPTR', 
    *           'STREAM-HANDLE' or 'LONGCHAR'.
    * @param    destination
    *           The appropriate object to write.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setOutputDestination(character mode, Object destination)
   {
      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusAttributeError("SET-OUTPUT-DESTINATION", writeStatus);
         return new logical(false);
      }

      final String writeMode;
      if (mode.isUnknown())
      {
         writeMode = "file";
      }
      else
      {
         writeMode = mode.getValue();
      }

      if ("file".equalsIgnoreCase(writeMode))
      {
         outputWrapper = new FileOutputStreamWrapper((String) destination);
      }
      else if ("stream".equalsIgnoreCase(writeMode))
      {
         LOG.warning("Stream is not supported");
         return new logical(false);
      }
      else if ("memptr".equalsIgnoreCase(writeMode))
      {
         outputWrapper = new MemptrOutputStream((memptr) destination);
      }
      else if ("stream-handle".equalsIgnoreCase(writeMode))
      {
         LOG.warning("Stream-handle is not supported");
         return new logical(false);
      }
      else if ("longchar".equalsIgnoreCase(writeMode))
      {
         outputWrapper = new LongcharOutputStream((longchar) destination);
      }
      else
      {
         invalidReadWriteModeError("SET-OUTPUT-DESTINATION");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Creates the XML document with the prolog information.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startDocument()
   {
      if (!WriteStatus.nonWritableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("START-DOCUMENT", writeStatus);
         return new logical(false);
      }

      if (outputWrapper == null)
      {
         String msg = "START-DOCUMENT was unable to open the output destination. " +
                      "No output destination was specified";
         ErrorManager.recordOrShowError(14610, msg, false, false, false);
      }

      XMLOutputFactory2 output = (XMLOutputFactory2) XMLOutputFactory2.newFactory();

      try
      {
         if (formatted)
         {
            writer = new IndentingXMLStreamWriter(output.createXMLStreamWriter(outputWrapper));
         }
         else
         {
            writer = output.createXMLStreamWriter(outputWrapper);
         }
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("START-DOCUMENT");
         return new logical(false);
      }

      try
      {
         if (standalone.isUnknown())
         {
            writer.writeStartDocument(encoding, version);
         }
         else
         {
            writeXMLDeclaration(writer, version, encoding, standalone.booleanValue());
         }
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("START-DOCUMENT");
         return new logical(false);
      }

      writeStatus = WriteStatus.SAX_WRITE_BEGIN;

      return new logical(true);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(Text name)
   {
      return startElement(name, new character(), null);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(String name)
   {
      return startElement(new character(name), new character(), null);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(Text name, handle handle)
   {
      return startElement(name, new character(), handle);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(String name, handle handle)
   {
      return startElement(new character(name), new character(), handle);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(Text name, Text uri)
   {
      return startElement(new character(name), uri, null);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(String name, Text uri)
   {
      return startElement(new character(name), uri, null);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(Text name, String uri)
   {
      return startElement(name, new character(uri), null);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(String name, String uri)
   {
      return startElement(new character(name), new character(uri), null);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(String name, Text uri, handle handle)
   {
      return startElement(new character(name), uri, handle);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(Text name, String uri, handle handle)
   {
      return startElement(name, new character(uri), handle);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(String name, String uri, handle handle)
   {
      return startElement(new character(name), new character(uri), handle);
   }

   /**
    * Adding the single attribute to the start tag in the XML document if applied for SAX writer
    * object. For SAX attributes object inserts attribute and its name into SAX attributes object.
    *
    * @param    name
    *           The attribute name to insert.
    * @param    value
    *           The attribute value to insert.
    * @param    uri
    *           The URI of the attribute to insert or '?' for unknown value.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical insertAttribute(Text name, Text value, Text uri)
   {
      if (!BaseDataType.elementsOfType(character.class, uri, name, value))
      {
         invalidArgumentError("INSERT-ATTRIBUTE");
         return new logical(false);
      }

      if (name.isUnknown() || value.isUnknown())
      {
         invalidArgumentError("INSERT-ATTRIBUTE");
         return new logical(false);
      }

      try
      {
         if (uri.isUnknown())
         {
            writer.writeAttribute(name.toStringMessage(), value.toStringMessage());
         }
         else
         {
            writer.writeAttribute(uri.toStringMessage(),
                                   name.toStringMessage(),
                                   value.toStringMessage());
         }
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("INSERT-ATTRIBUTE");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Starts the XML node according to the specified name, namespace URI and handle of the SAX
    * attribute object.
    * 
    * @param    name
    *           The name of the element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    attr
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical startElement(Text name, Text uri, handle attr)
   {
      if (!BaseDataType.elementsOfType(character.class, uri, name))
      {
         invalidArgumentError("START-ELEMENT");
         return new logical(false);
      }

      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("START-ELEMENT", writeStatus);
         return new logical(false);
      }

      if (name.isUnknown())
      {
         invalidArgumentError("START-ELEMENT");
         return new logical(false);
      }
      if (attr != null && !attr.isUnknown() && !handle.isHandleTypeOf(attr, SaxAttributes.class))
      {
         invalidArgumentError("START-ELEMENT");
         return new logical(false);
      }

      Attributes attributes = ((SaxAttributesImpl)attr.getResource()).getAttributes();

      try
      {
         String uriStr = uri.isUnknown() ? null : uri.toStringMessage();

         writer.writeStartElement(uriStr, name.toStringMessage());
         writeStatus = WriteStatus.SAX_WRITE_TAG;
         for(int i = 0; i < attributes.getLength(); i++)
         {
            logical res = insertAttribute(attributes.getLocalName(i), attributes.getValue(i));
            if (!res.booleanValue())
            {
               return res;
            }
         }
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-ELEMENT");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Adds CDATA block to the XML document crating by SAX-Writer object.
    * 
    * @param    cdataBlock
    *           The data to be written with CDATA block.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeCdata(String cdataBlock)
   {
      return writeCdata(new character(cdataBlock));
   }

   /**
    * Adds CDATA block to the XML document crating by SAX-Writer object.
    * 
    * @param    cdataBlock
    *           The data to be written with CDATA block.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeCdata(Text cdataBlock)
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-CDATA", writeStatus);
         return new logical(false);
      }

      if (cdataBlock.isUnknown())
      {
         invalidArgumentError("WRITE-CDATA");
         return new logical(false);
      }

      try
      {
         writer.writeCData(cdataBlock.toStringMessage());
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-CDATA");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Adds character data to the XML document crating by SAX-Writer object.
    * 
    * @param    chardata
    *           The character or longchar data to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeCharacters(String chardata)
   {
      return writeCharacters(new character(chardata));
   }

   /**
    * Adds character data to the XML document crating by SAX-Writer object.
    * 
    * @param    chardata
    *           The character or longchar data to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeCharacters(Text chardata)
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-CHARACTERS", writeStatus);
         return new logical(false);
      }

      if (chardata.isUnknown())
      {
         invalidArgumentError("WRITE-CHARACTERS");
         return new logical(false);
      }

      try
      {
         writer.writeCharacters(chardata.toStringMessage());
         writeStatus = WriteStatus.SAX_WRITE_CONTENT;
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-CHARACTERS");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, Text chardata)
   {
      return writeDataElement(name, chardata, new character(), null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, Text chardata)
   {
      return writeDataElement(new character(name), chardata, new character(), null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, String chardata)
   {
      return writeDataElement(name, new character(chardata), new character(), null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, String chardata)
   {
      return writeDataElement(new character(name), new character(chardata), new character(),
                              null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, Text chardata, Text uri)
   {
      return writeDataElement(name, chardata, uri, null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, Text chardata, Text uri)
   {
      return writeDataElement(new character(name), chardata, uri, null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, String chardata, Text uri)
   {
      return writeDataElement(name, new character(chardata), uri, null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, String chardata, Text uri)
   {
      return writeDataElement(new character(name), new character(chardata), uri, null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, Text chardata, String uri)
   {
      return writeDataElement(name, chardata, new character(uri), null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, Text chardata, String uri)
   {
      return writeDataElement(new character(name), chardata, new character(uri), null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, String chardata, String uri)
   {
      return writeDataElement(name, new character(chardata), new character(uri), null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, String chardata, String uri)
   {
      return writeDataElement(new character(name), new character(chardata),
                              new character(uri), null);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, Text chardata, handle handle)
   {
      return writeDataElement(name, chardata, new character(), handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, Text chardata, handle handle)
   {
      return writeDataElement(new character(name), chardata, new character(), handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, String chardata, handle handle)
   {
      return writeDataElement(name, new character(chardata), new character(), handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, String chardata, handle handle)
   {
      return writeDataElement(new character(name), new character(chardata), new character(),
                              handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, Text chardata, Text uri, handle handle)
   {
      return writeDataElement(new character(name), chardata, uri, handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, String chardata, Text uri, handle handle)
   {
      return writeDataElement(name, new character(chardata), uri, handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, Text chardata, String uri, handle handle)
   {
      return writeDataElement(name, chardata, new character(uri), handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, String chardata, Text uri, handle handle)
   {
      return writeDataElement(new character(name), new character(chardata), uri, handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, Text chardata, String uri, handle handle)
   {
      return writeDataElement(new character(name), chardata, new character(uri), handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, String chardata, String uri, handle handle)
   {
      return writeDataElement(name, new character(chardata), new character(uri), handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    handle
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(String name, String chardata, String uri, handle handle)
   {
      return writeDataElement(new character(name), new character(chardata),
                              new character(uri), handle);
   }

   /**
    * Creates complete XML node with a given SAX-Writer object.
    * 
    * @param    name
    *           The name of the element.
    * @param    chardata
    *           The XML text to write for the given element.
    * @param    uri
    *           The namespace URI of the given element.
    * @param    attr
    *           The handle of the SAX-Attribute object to be added into XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeDataElement(Text name, Text chardata, Text uri, handle attr)
   {
      if (!BaseDataType.elementsOfType(character.class, uri, name))
      {
         invalidArgumentError("WRITE-DATA-ELEMENT");
         return new logical(false);
      }

      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-DATA-ELEMENT", writeStatus);
         return new logical(false);
      }

      if (name.isUnknown())
      {
         invalidArgumentError("WRITE-DATA-ELEMENT");
         return new logical(false);
      }

      if (attr != null && !attr.isUnknown() && !handle.isHandleTypeOf(attr, SaxAttributes.class))
      {
         invalidArgumentError("WRITE-DATA-ELEMENT");
         return new logical(false);
      }

      Attributes attributes = ((SaxAttributesImpl)attr.getResource()).getAttributes();

      try
      {
         String uriStr = uri.isUnknown() ? null : uri.toStringMessage();

         writer.writeEmptyElement(uriStr, name.toStringMessage());
         writeStatus = WriteStatus.SAX_WRITE_TAG;
         for(int i = 0; i < attributes.getLength(); i++)
         {
            logical res = insertAttribute(attributes.getLocalName(i), attributes.getValue(i));
            if (!res.booleanValue())
            {
               return res;
            }
         }
         if (!chardata.isUnknown())
         {
            writer.writeCharacters(chardata.toStringMessage());
         }
         writer.writeEndElement();
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-EMPTY-ELEMENT");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Adds character data to the XML document crating by SAX-Writer object.
    * 
    * @param    chardata
    *           The character or longchar data to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeFragment(String chardata)
   {
      return writeFragment(new character(chardata));
   }

   /**
    * Adds character data to the XML document crating by SAX-Writer object.
    * 
    * @param    chardata
    *           The character or longchar data to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeFragment(Text chardata)
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-FRAGMENT", writeStatus);
         return new logical(false);
      }

      if (chardata.isUnknown())
      {
         invalidArgumentError("WRITE-FRAGMENT");
         return new logical(false);
      }

      String content = chardata.toStringMessage();

      return writeCharacterData(content);
   }

   private logical writeCharacterData(String content)
   {
      try
      {
         writer.writeCharacters(content);
         writeStatus = WriteStatus.SAX_WRITE_CONTENT;
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-FRAGMENT");
         return new logical(false);
      }
      return new logical(true);
   }

   /**
    * Adds character data to the XML document crating by SAX-Writer object.
    * 
    * @param    xhandle
    *           The valid X-Nodereference handle containing XML text.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeFragment(handle xhandle)
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-FRAGMENT", writeStatus);
         return new logical(false);
      }

      if (!XEntityImpl.isXNodeHandle(xhandle) || !XEntityImpl.isXNodeHandleInitialized(xhandle))
      {
         invalidArgumentError("WRITE-FRAGMENT");
         return new logical(false);
      }

      Node node = ((XEntityImpl) xhandle.get()).getNode();

      String nodeXML;
      try
      {
         nodeXML = toString(node);
      }
      catch (TransformerException ignored)
      {
         invalidArgumentError("WRITE-FRAGMENT");
         return new logical(false);
      }


      return writeCharacterData(nodeXML);
   }
   
   /**
    * Closes the open stream and resets the SAX-Writer object to the default state.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical reset()
   {
      return new logical(true);
   }

   /**
    * Adds comment to the XML document crating by SAX-Writer object.
    * 
    * @param    comment
    *           The character or longchar text comment to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeComment(String comment)
   {
      return writeComment(new character(comment));
   }

   /**
    * Adds comment to the XML document crating by SAX-Writer object.
    * 
    * @param    comment
    *           The character or longchar text comment to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeComment(Text comment)
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-COMMENT", writeStatus);
         return new logical(false);
      }

      if (comment.isUnknown())
      {
         invalidArgumentError("WRITE-COMMENT");
         return new logical(false);
      }

      try
      {
         writer.writeComment(comment.toStringMessage());
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-COMMENT");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(Text name)
   {
      return writeEmptyElement(name, new character(), null);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(String name)
   {
      return writeEmptyElement(new character(name), new character(), null);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(Text name, Text uri)
   {
      return writeEmptyElement(name, uri, null);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(String name, Text uri)
   {
      return writeEmptyElement(new character(name), uri, null);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(Text name, String uri)
   {
      return writeEmptyElement(name, new character(uri), null);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(String name, String uri)
   {
      return writeEmptyElement(new character(name), new character(uri), null);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    attr
    *           The handle of the SAX-Attributes object to be used.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(Text name, handle attr)
   {
      return writeEmptyElement(name, new character(), attr);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    attr
    *           The handle of the SAX-Attributes object to be used.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(String name, handle attr)
   {
      return writeEmptyElement(new character(name), new character(), attr);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * @param    attr
    *           The handle of the SAX-Attributes object to be used.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(String name, Text uri, handle attr)
   {
      return writeEmptyElement(new character(name), uri, attr);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * @param    attr
    *           The handle of the SAX-Attributes object to be used.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(Text name, String uri, handle attr)
   {
      return writeEmptyElement(name, new character(uri), attr);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * @param    attr
    *           The handle of the SAX-Attributes object to be used.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(String name, String uri, handle attr)
   {
      return writeEmptyElement(new character(name), new character(uri), attr);
   }

   /**
    * Creates an empty node in a given SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar element name to be written.
    * @param    uri
    *           The character or longchar element URI to be written.
    * @param    attr
    *           The handle of the SAX-Attributes object to be used.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEmptyElement(Text name, Text uri, handle attr)
   {
      if (!BaseDataType.elementsOfType(character.class, uri, name))
      {
         invalidArgumentError("WRITE-EMPTY-ELEMENT");
         return new logical(false);
      }

      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-EMPTY-ELEMENT", writeStatus);
         return new logical(false);
      }

      if (name.isUnknown())
      {
         invalidArgumentError("WRITE-EMPTY-ELEMENT");
         return new logical(false);
      }

      if (attr != null && !attr.isUnknown() && !handle.isHandleTypeOf(attr, SaxAttributes.class))
      {
         invalidArgumentError("WRITE-EMPTY-ELEMENT");
         return new logical(false);
      }

      Attributes attributes = ((SaxAttributesImpl)attr.getResource()).getAttributes();

      try
      {
         String uriStr = uri.isUnknown() ? null : uri.toStringMessage();

         writer.writeEmptyElement(uriStr, name.toStringMessage());
         writeStatus = WriteStatus.SAX_WRITE_TAG;
         for(int i = 0; i < attributes.getLength(); i++)
         {
            logical res = insertAttribute(attributes.getLocalName(i), attributes.getValue(i));
            if (!res.booleanValue())
            {
               return res;
            }
         }
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-EMPTY-ELEMENT");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Adds an entity referenc to the XML document associated with given SAX-Writer object.
    * 
    * @param    value
    *           The character or longchar value to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEntityRef(String value)
   {
      return writeEntityRef(new character(value));
   }

   /**
    * Adds an entity referenc to the XML document associated with given SAX-Writer object.
    * 
    * @param    value
    *           The character or longchar value to be written.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeEntityRef(Text value)
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-ENTITY-REF", writeStatus);
         return new logical(false);
      }

      if (value.isUnknown())
      {
         invalidArgumentError("WRITE-ENTITY-REF");
         return new logical(false);
      }

      try
      {
         writer.writeEntityRef(value.toStringMessage());
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-ENTITY-REF");
         return new logical(false);
      }
      return new logical(true);
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(Text name, Text systemId)
   {
      return writeExternalDtd(name, systemId, new character());
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(String name, Text systemId)
   {
      return writeExternalDtd(new character(name), systemId, new character());
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(Text name, String systemId)
   {
      return writeExternalDtd(name, new character(systemId), new character());
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(String name, String systemId)
   {
      return writeExternalDtd(new character(name), new character(systemId), new character());
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(String name, Text systemId, Text publicId)
   {
      return writeExternalDtd(new character(name), systemId, publicId);
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(Text name, String systemId, Text publicId)
   {
      return writeExternalDtd(name, new character(systemId), publicId);
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(String name, String systemId, Text publicId)
   {
      return writeExternalDtd(new character(name), new character(systemId), publicId);
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(Text name, Text systemId, String publicId)
   {
      return writeExternalDtd(name, systemId, new character(publicId));
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(String name, Text systemId, String publicId)
   {
      return writeExternalDtd(new character(name), systemId, new character(publicId));
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(Text name, String systemId, String publicId)
   {
      return writeExternalDtd(name, new character(systemId), new character(publicId));
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(String name, String systemId, String publicId)
   {
      return writeExternalDtd(new character(name), new character(systemId),
                              new character(publicId));
   }

   /**
    * Adds an external DTD reference to the XML document associated with a SAX-Writer object.
    * 
    * @param    name
    *           The character or longchar name of the document root node.
    * @param    systemId
    *           The character or longchar system ID of the DTD.
    * @param    publicId
    *           The character or longchar public ID of the DTD.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeExternalDtd(Text name, Text systemId, Text publicId)
   {
      if (writeStatus != WriteStatus.SAX_WRITE_BEGIN)
      {
         invalidStatusMethodError("WRITE-EXTERNAL-DTD", writeStatus);
         return new logical(false);
      }

      if (name.isUnknown())
      {
         invalidArgumentError("WRITE-EXTERNAL-DTD");
         return new logical(false);
      }

      if (systemId.isUnknown())
      {
         invalidArgumentError("WRITE-EXTERNAL-DTD");
         return new logical(false);
      }

      StringBuilder xmlDeclarationBuilder = new StringBuilder();
      xmlDeclarationBuilder.append("<!DOCTYPE ");
      xmlDeclarationBuilder.append(name.toStringMessage());
      xmlDeclarationBuilder.append(' ');
      xmlDeclarationBuilder.append(systemId.toStringMessage());
      if (!publicId.isUnknown())
      {
         xmlDeclarationBuilder.append(" \"");
         xmlDeclarationBuilder.append(publicId.toStringMessage());
         xmlDeclarationBuilder.append('"');
      }
      xmlDeclarationBuilder.append(">\n");

      try
      {
         writer.writeDTD(xmlDeclarationBuilder.toString());
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-EXTERNAL-DTD");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Creates a processing instruction node in XML document for given SAX-Writer object.
    * 
    * @param    target
    *           The character or longchar target of the processing instruction.
    * @param    data
    *           The character or longchar data associated with the processing instruction.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeProcessingInstruction(String target, Text data)
   {
      return writeProcessingInstruction(new character(target), data);
   }

   /**
    * Creates a processing instruction node in XML document for given SAX-Writer object.
    * 
    * @param    target
    *           The character or longchar target of the processing instruction.
    * @param    data
    *           The character or longchar data associated with the processing instruction.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeProcessingInstruction(Text target, String data)
   {
      return writeProcessingInstruction(target, new character(data));
   }

   /**
    * Creates a processing instruction node in XML document for given SAX-Writer object.
    * 
    * @param    target
    *           The character or longchar target of the processing instruction.
    * @param    data
    *           The character or longchar data associated with the processing instruction.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeProcessingInstruction(String target, String data)
   {
      return writeProcessingInstruction(new character(target), new character(data));
   }

   /**
    * Creates a processing instruction node in XML document for given SAX-Writer object.
    * 
    * @param    target
    *           The character or longchar target of the processing instruction.
    * @param    data
    *           The character or longchar data associated with the processing instruction.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical writeProcessingInstruction(Text target, Text data)
   {
      if (!WriteStatus.writableStatuses.contains(writeStatus))
      {
         invalidStatusMethodError("WRITE-PROCESSING-INSTRUCTION", writeStatus);
         return new logical(false);
      }

      if (target.isUnknown())
      {
         invalidArgumentError("WRITE-PROCESSING-INSTRUCTION");
         return new logical(false);
      }

      if (data.isUnknown())
      {
         invalidArgumentError("WRITE-PROCESSING-INSTRUCTION");
         return new logical(false);
      }

      try
      {
         writer.writeProcessingInstruction(target.toStringMessage(), data.toStringMessage());
      }
      catch (XMLStreamException ignored)
      {
         writerOutputError("WRITE-PROCESSING-INSTRUCTION");
         return new logical(false);
      }

      return new logical(true);
   }
   
   /**
    * Check if this resource supports the NAME attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasName()
   {
      return false;
   }

   /**
    * This method is used to serialize Node from DOM to string.
    *
    * @param    node
    *           Node instance from DOM model.
    *
    * @return   Text representation of the node.
    *
    * @throws   TransformerFactoryConfigurationError If an error occurs writing the identifier.
    * @throws   TransformerException If an error occurs writing the identifier.
    */
   static String toString(Node node)
   throws TransformerFactoryConfigurationError,
          TransformerException
   {
      StreamResult xmlOutput = new StreamResult(new StringWriter());
      Transformer transformer = TransformerFactory.newInstance().newTransformer();
      transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
      transformer.transform(new DOMSource(node), xmlOutput);
      return xmlOutput.getWriter().toString();
   }

   /**
    * This method is used to write the XML document declaration at the beginning of an XML
    * document. This version is used when the standalone attribute is to be written as the
    * XMLStreamWriter provides no way for writing this attribute.
    *
    * @param    outputWriter
    *           XMLStreamWriter instance.
    * @param    xmlVersion
    *           Xml version which will specify in xml header.
    * @param    encodingName
    *           encoding which will specify in xml header.
    * @param    standalone
    *           standalone flag.
    * @throws   XMLStreamException
    *           If an error occurs writing the identifier.
    */
   public static void writeXMLDeclaration(XMLStreamWriter outputWriter,
                                          String xmlVersion,
                                          String encodingName,
                                          boolean standalone)
   throws XMLStreamException
   {
      if (xmlVersion == null || xmlVersion.trim().isEmpty())
      {
         xmlVersion = DEFAULT_XML_VERSION;
      }
      StringBuilder xmlDeclarationBuilder = new StringBuilder();
      xmlDeclarationBuilder.append("<?xml version=\"");
      xmlDeclarationBuilder.append(xmlVersion);
      xmlDeclarationBuilder.append('"');
      if (encodingName != null)
      {
         xmlDeclarationBuilder.append(" encoding=\"");
         xmlDeclarationBuilder.append(encodingName);
         xmlDeclarationBuilder.append('"');
      }
      xmlDeclarationBuilder.append(" standalone=\"");
      xmlDeclarationBuilder.append(standalone ? "yes" : "no");
      xmlDeclarationBuilder.append("\"?>\n");
      //Work around to avoid the angle brackets getting escaped.
      outputWriter.writeDTD(xmlDeclarationBuilder.toString());
   }

   /**
    * Special wrapper to wrap write to file in lazy mode. It's means that first try to write
    * will check if file is available.
    */
   private static class FileOutputStreamWrapper
   extends OutputStream
   {
      /** File name.*/
      private final String fileName;

      /** Lazy initialization flag.*/
      private volatile boolean isStarted;

      /** Buffered output Stream.*/
      private Stream stream;

      /**
       * Constructor accepting a file name, which will read after first call.
       *
       * @param    source
       *           A file name.
       */
      private FileOutputStreamWrapper(String source)
      {
         isStarted = false;
         fileName = source;
      }

      @Override
      public void write(int b)
              throws IOException
      {
         if (!isStarted)
         {
            init(true);
         }
         stream.writeByte((byte) b);
      }

      @Override
      public void close()
      throws IOException
      {
         if (stream != null)
         {
            stream.close();
         }
      }

      /**
       * Initialization method which creates stream wrapper.
       *
       * @param    canWrite
       *           if {@code true} than file will open in read/write mode, otherwise only in
       *           read mode.
       */
      private void init(boolean canWrite) throws IOException
      {
         try
         {
            stream = StreamFactory.openFileStream(fileName, canWrite, false);
            if (stream == null)
            {
               throw new IOException("An exception occurred!");
            }
         }
         catch (ErrorConditionException e)
         {
            throw new IOException("An exception occurred!", e);
         }
         isStarted = true;
      }
   }

   /**
    * Special wrapper to wrap write to longchar in lazy mode. It's means that first try to write
    * will check if longchar is available.
    */
   private static class LongcharOutputStream
   extends OutputStream
   {
      /** A valid longchar object to write.*/
      private longchar lchar;

      /** Buffered output Stream.*/
      private StringWriter buffer = new StringWriter();

      /**
       * Constructor accepting a longchar, which will read after first call.
       *
       * @param    source
       *           A valid longchar object.
       */
      private LongcharOutputStream(longchar source)
      {
         lchar = source;
         lchar.assign("");
      }

      @Override
      public void write(int b) throws IOException
      {
         buffer.write(b);
      }

      @Override
      public void flush() throws IOException
      {
         String buf = buffer.getBuffer().toString();
         buffer.getBuffer().setLength(0);
         lchar.assign(TextOps.concat(lchar, buf));
      }

      @Override
      public void close()
      throws IOException
      {
         flush();
         buffer.close();
      }
   }

   /**
    * Special wrapper to wrap write to memptr in lazy mode. It's means that first try to write
    * will check if memptr is available.
    */
   private static class MemptrOutputStream
   extends OutputStream
   {
      /** A valid memptr */
      private final memptr mptr;

      /** Current read cursor position, started from index 1.*/
      private final AtomicLong position = new AtomicLong();

      /**
       * Constructor accepting a memptr to write.
       *
       * @param    source
       *           A valid memptr object.
       */
      private MemptrOutputStream(memptr source)
      {
         mptr = source;
         mptr.setLength(1); //reset memptr
         position.set(1);
      }

      @Override
      public void write(int b) throws IOException
      {
         try
         {
            mptr.setByte(b, position.getAndIncrement());
         }
         catch (ErrorConditionException e)
         {
            throw new IOException(e);
         }
      }
   }
}
